const express = require("express");
const Game = require("./game.js");

class Empathy extends Game {
  constructor(id) {
    super(id);
    this.state = {
      prompts: [],
      active_prompt: null,
      players_answered: 0,
      ambiguities: null,
      players_judged: 0,
      scores: null
    };
    this.answers = [];
    this.next_prompt_id = 1;
    this.equivalencies = {};
  }

  reset() {

    /* Before closing out the current round, we accumulate that score
     * for each player into their runnning total. */
    for (let score of this.state.scores.scores) {
      const player = this.players.find(p => p.name === score.player);
      if (player.score)
        player.score += score.score;
      else
        player.score = score.score;

      /* And broadcast that new score out. */
      this.broadcast_event('player-update', player.info_json());
    }

    /* Now that we're done with the active prompt, we remove it from
     * the list of prompts and also remove any prompts that received
     * no votes. This keeps the list of prompts clean.
     */
    const active_id = this.state.active_prompt.id;
    this.state.prompts =
      this.state.prompts.filter(
        p => p.id !== active_id && p.votes.length > 0
      );

    this.state.active_prompt = null;
    this.state.players_answered = 0;
    this.state.ambiguities = 0;
    this.state.players_judged = 0;
    this.state.scores = null;

    this.answers = [];
    this.equivalencies = {};

    this.broadcast_event_object('game-state', this.state);
  }

  add_prompt(items, prompt_string) {
    const prompt = new Prompt(this.next_prompt_id, items, prompt_string);
    this.next_prompt_id++;

    this.state.prompts.push(prompt);

    this.broadcast_event_object('prompt', prompt);

    return prompt;
  }

  /* Returns true if vote toggled, false for player or prompt not found */
  toggle_vote(prompt_id, session_id) {
    const player = this.players_by_session[session_id];

    const prompt = this.state.prompts.find(p => p.id === prompt_id);
    if (! prompt || ! player)
      return false;

    prompt.toggle_vote(player.name);

    this.broadcast_event_object('prompt', prompt);

    return true;
  }

  /* Returns true on success, false for prompt not found. */
  start(prompt_id) {
    const prompt = this.state.prompts.find(p => p.id === prompt_id);
    if (! prompt)
      return false;

    /* Ignore any start request that comes in while a prompt is
     * already being played. */
    if (this.state.active_prompt)
      return false;

    this.state.active_prompt = prompt;

    this.broadcast_event_object('start', prompt);

    return true;
  }

  receive_answer(prompt_id, session_id, answers) {
    const player = this.players_by_session[session_id];
    if (! player)
      return { valid: false, message: "Player not found" };

    const prompt = this.state.prompts.find(p => p.id === prompt_id);
    if (! prompt)
      return { valid: false, message: "Prompt not found" };

    if (prompt !== this.state.active_prompt)
      return { valid: false, message: "Prompt no longer active" };

    /* Save the complete answers for our own use later. */
    this.answers.push({
      player: player,
      answers: answers
    });

    /* And notify players how many players have answered. */
    this.state.players_answered++;
    this.broadcast_event_object('answered', this.state.players_answered);

    return { valid: true };
  }

  perform_judging() {
    const word_map = {};

    for (let a of this.answers) {
      for (let word of a.answers) {
        const key = this.canonize(word);
        word_map[key] = word;
      }
    }

    this.state.ambiguities = Object.values(word_map);

    this.broadcast_event_object('ambiguities', this.state.ambiguities);
  }

  receive_judging(prompt_id, session_id, word_groups) {
    const player = this.players_by_session[session_id];
    if (! player)
      return { valid: false, message: "Player not found" };

    const prompt = this.state.prompts.find(p => p.id === prompt_id);
    if (! prompt)
      return { valid: false, message: "Prompt not found" };

    if (prompt !== this.state.active_prompt)
      return { valid: false, message: "Prompt no longer active" };

    /* Each player submits some number of groups of answers that
     * should be considered equivalent. The server expands that into
     * the set of pair-wise equivalencies that are expressed. The
     * reason for that is so that the server can determine which
     * pair-wise equivalencies have majority support.
     */
    for (let group of word_groups) {

      for (let i = 0; i < group.length - 1; i++) {
        for (let j = i + 1; j < group.length; j++) {
          let eq = [group[i], group[j]];

          /* Put the two words into a reliable order so that we don't
           * miss a pair of equivalent equivalencies just because they
           * happen to be in the opposite order. */
          if (eq[0].localeCompare(eq[1]) > 0) {
            eq = [group[j], group[i]];
          }

          const exist = this.equivalencies[`${eq[0]}:${eq[1]}`];
          if (exist) {
            exist.count++;
          } else {
            this.equivalencies[`${eq[0]}:${eq[1]}`] = {
              count: 1,
              words: eq
            };
          }

        }
      }
    }

    /* And notify players how many players have completed judging. */
    this.state.players_judged++;
    this.broadcast_event_object('judged', this.state.players_judged);

    return { valid: true };
  }

  canonize(word) {
    return word.toLowerCase();
  }

  compute_scores() {
    const word_submitters = {};
    const scores = [];

    /* Perform a (non-strict) majority ruling on equivalencies,
     * dropping all that didn't get enough votes. */
    const quorum = Math.floor((this.players.length + 1)/2);
    const agreed_equivalencies = Object.values(this.equivalencies).filter(
      eq => eq.count >= quorum);

    /* And with that agreed set of equivalencies, construct the full
     * groups. */
    const word_groups = {};

    for (let e of agreed_equivalencies) {
      let group = word_groups[e.words[0]];
      if (! group)
        group = word_groups[e.words[1]];
      if (! group)
        group = { words: [], players: new Set()};

      if (! word_groups[e.words[0]]) {
        word_groups[e.words[0]] = group;
        group.words.push(e.words[0]);
      }

      if (! word_groups[e.words[1]]) {
        word_groups[e.words[1]] = group;
        group.words.push(e.words[1]);
      }
    }

    /* Now go through answers from each player and add the player name
     * to the set corresponding to each word group. */
    for (let a of this.answers) {
      for (let word of a.answers) {
        /* If there's no group yet, this is a singleton word. */
        if (word_groups[word]) {
          word_groups[word].players.add(a.player.name);
        } else {
          const group = { words: [word], players: new Set() };
          group.players.add(a.player.name);
          word_groups[word] = group;
        }
      }
    }

    /* Finally, go through the answers one more time, this time taking
     * the length of the player set as a score for the player's
     * submission of that word. */
    for (let a of this.answers) {
      let score = 0;
      for (let word of a.answers) {
        score += word_groups[word].players.size;
      }
      scores.push({
        player: a.player.name,
        score: score
      });
    }

    scores.sort((a,b) => {
      return b.score - a.score;
    });

    /* Put the word groups into a form the client can consume.
     *
     * Most significantly, we only want one entry for each group (even
     * though our current "word_groups" object has a property for each
     * word considered equivalent).
     */
    const words_submitted = Object.entries(word_groups).filter(
      group => group[0] === group[1].words[0]).map(
        group => {
          return {
            word: group[1].words.join('/'),
            players: Array.from(group[1].players)
          };
        }
      );

    words_submitted.sort((a,b) => {
      return b.players.length - a.players.length;
    });

    /* Put this round's scores into the game state object so it will
     * be sent to any new clients that join. */
    this.state.scores = {
      scores: scores,
      words: words_submitted
    };

    /* And broadcast the scores to all connected clients. */
    this.broadcast_event_object('scores', this.state.scores);
  }
}

Empathy.router = express.Router();
const router = Empathy.router;

class Prompt {
  constructor(id, items, prompt) {
    this.id = id;
    this.items = items;
    this.prompt = prompt;
    this.votes = [];
  }

  toggle_vote(player_name) {
    if (this.votes.find(v => v === player_name))
      this.votes = this.votes.filter(v => v !== player_name);
    else
      this.votes.push(player_name);
  }
}

router.post('/prompts', (request, response) => {
  const game = request.game;

  prompt = game.add_prompt(request.body.items, request.body.prompt);

  response.json({ id: prompt.id});
});

router.post('/vote/:prompt_id([0-9]+)', (request, response) => {
  const game = request.game;
  const prompt_id = parseInt(request.params.prompt_id, 10);

  if (game.toggle_vote(prompt_id, request.session.id))
    response.send('');
  else
    response.sendStatus(404);
});

router.post('/start/:prompt_id([0-9]+)', (request, response) => {
  const game = request.game;
  const prompt_id = parseInt(request.params.prompt_id, 10);

  if (game.start(prompt_id))
    response.send('');
  else
    response.sendStatus(404);
});

router.post('/answer/:prompt_id([0-9]+)', (request, response) => {
  const game = request.game;
  const prompt_id = parseInt(request.params.prompt_id, 10);

  const result = game.receive_answer(prompt_id,
                                     request.session.id,
                                     request.body.answers);
  response.json(result);

  if (game.state.players_answered >= game.players.length)
    game.perform_judging();
});

router.post('/judging/:prompt_id([0-9]+)', (request, response) => {
  const game = request.game;
  const prompt_id = parseInt(request.params.prompt_id, 10);

  const result = game.receive_judging(prompt_id,
                                      request.session.id,
                                      request.body.word_groups);
  response.json(result);

  if (game.state.players_judged >= game.players.length)
    game.compute_scores();
});

router.post('/reset', (request, response) => {
  const game = request.game;
  game.reset();
});

Empathy.meta = {
  name: "Empathy",
  identifier: "empathy",
};

exports.Game = Empathy;
